#include "stdafx.h"

const int MINGENSAMPLES=10;

//----- Global Array of 'discovered' cards.
int g_iCardCount=-1;				
int NiCardCount() { 
	return g_iCardCount;
	}
	
CNICard * g_aCards;					
CNICard * NiCard(int iCard){
	return (iCard<0||iCard>g_iCardCount) ? NULL : &(g_aCards[iCard]) ;
	}	

//-----------------------------------------------------------------------------
bool CFeedback :: LoadFromFile( CFeedback ** ppFeedback, FILE * h, int iVer ){	
	int iFeedbackType;
	//-- read in feedback type then construct one of the descendant classes 
	fread( &iFeedbackType, sizeof(int), 1, h );
	if( iFeedbackType==FEEDBACKTYPE_PID ) 
		*ppFeedback = (CFeedback *) new CPID;
	else 
		return false;
				
	if( *ppFeedback == NULL 
			|| ! FileLoadString( h, &((*ppFeedback)->m_cName) ) 
			|| ! (*ppFeedback)->LoadHelper(h, iVer) ) 
		return false;
	return true;
	}	
	
bool CFeedback :: SaveToFile( FILE * h ){	
	return fwrite( &m_iFeedbackType, sizeof(int), 1, h )==1 
		&& FileSaveString( h, m_cName.ConstBuffer() )
		&& SaveHelper(h);	
	}	
	
int CFeedback :: SaveSize(){	
	return sizeof( m_iFeedbackType ) + 5 + m_cName.Length() + SaveSizeHelper();
	}	
	
CFeedback * CFeedback :: NewCopy() { 
	if( m_iFeedbackType==FEEDBACKTYPE_PID ) {
		CPID * pPID = new CPID( *(CPID *)this ); 
		pPID->m_adQueueValue=NULL;
		pPID->m_aiQueueTime=NULL;
		pPID->QueueAllocate( pPID->m_iQueueSize );
		return pPID;
		}
	return NULL;
	}
	
	
//-----------------------------------------------------------------------------
CPID :: CPID() {
	m_iFeedbackType = FEEDBACKTYPE_PID;
	m_dKp=1.0;
	m_dKi=1.0;
	m_dKd=1.0;
	m_iQueueSize=10;
	m_adQueueValue=NULL;
	m_aiQueueTime=NULL;
	QueueAllocate(PID_QUE_DEFAULT_SIZE);
	m_iQueueLast=-1;
	}
	
CPID :: ~CPID() { 
	QueueAllocate(0);
	}

void CPID :: Set( double dDeltaT, double dKp, double dKi, double dKd, int iQueueSize, int iAIIndex, int iAOIndex ) {
	m_dDeltaT=dDeltaT;
	m_dKp=dKp;
	m_dKi=dKi;
	m_dKd=dKd;
	QueueAllocate( iQueueSize );
	m_iInputIndex=iAIIndex;
	m_iOutputIndex=iAOIndex;
	m_iQueueLast=-1;
	}

bool CPID :: QueueAllocate( int iQueueSize ) { 
	m_iQueueSize=iQueueSize;
	m_dQueueSum=0.0;
	m_iQueueLast=-1;
	if( m_adQueueValue!=NULL )
		delete [] m_adQueueValue;
	if( m_aiQueueTime!=NULL ) 
		delete [] m_aiQueueTime;
	if( iQueueSize > 0 ) { 
		m_adQueueValue = new double[iQueueSize];
		m_aiQueueTime = new int[iQueueSize];
		if( m_adQueueValue==NULL || m_aiQueueTime==NULL ) {
			if( m_adQueueValue != NULL ) 
				delete [] m_adQueueValue;
			return false;
			}
		}
	else { 
		m_adQueueValue = NULL;
		m_aiQueueTime = NULL;
		}
	return true;
	}

double CPID :: CalculateOutput( double dSetpoint, double dInput, int iSample ) {
	/*
	#ifdef _DEBUG
	char bufODS[1024];
	sprintf( bufODS, "NIDAQ CalculateOutput( %f, %f, %d)\n", dSetpoint, dInput, iSample );
	OutputDebugString( bufODS );
	#endif
	*/

	double dErr = dSetpoint-dInput;
	//-- maintain the queue for integral window 
	if( m_iQueueLast==-1 ) {	
		//-- Initialize the queue - start of new queue 
		for( int i=0; i<m_iQueueSize; ++i ) {
			m_adQueueValue[i]=dErr;
			m_aiQueueTime[i]=iSample;
			}
		m_iQueueLast=0;
		m_dQueueSum=dErr*m_iQueueSize;
		}
	else if( ++m_iQueueSummed >= PID_QUE_RECALC_INTERVAL ) { 
		//-- periodically refresh the sum to avoid accumulated floating pt. errors (although likely they would cancel).
		m_iQueueSummed=0;
		m_dQueueSum=0.0;
		for( int i=0; i<m_iQueueSize; ++i )
			m_dQueueSum+=m_adQueueValue[i];
		}
		
	//-- add to the queue 
	int iPrevious = m_iQueueLast;
	m_iQueueLast = (m_iQueueLast+1)%m_iQueueSize;
	double dDeltaT = m_dDeltaT * (iSample-m_adQueueValue[m_iQueueLast]); // get time diff prior to overwrite
	m_dQueueSum += dErr-m_adQueueValue[m_iQueueLast];	// Sum : add new and remove oldest 
	m_adQueueValue[m_iQueueLast]=dErr;
	m_aiQueueTime[m_iQueueLast]=iSample;

	//-- P I D - Proportional Integral Differential components
	return m_dKp*dErr 
		+ m_dKi*m_dQueueSum/dDeltaT 
		+ m_dKd*(dErr-m_adQueueValue[iPrevious]);
	}

bool CPID :: LoadHelper( FILE * h, int iVer ) {
	return fread( &m_dKp, sizeof(double), 1, h )==1
		&& fread( &m_dKi, sizeof(double), 1, h )==1
		&& fread( &m_dKd, sizeof(double), 1, h )==1
		&& fread( &m_iQueueSize, sizeof(int), 1, h )==1
		&& QueueAllocate( m_iQueueSize );
	}

bool CPID :: SaveHelper( FILE * h ) { 
	return fwrite( &m_dKp, sizeof(double), 1, h )==1
		&& fwrite( &m_dKi, sizeof(double), 1, h )==1
		&& fwrite( &m_dKd, sizeof(double), 1, h )==1
		&& fwrite( &m_iQueueSize, sizeof(int), 1, h )==1;
	}

int CPID :: SaveSizeHelper() { 
	return 3*sizeof(double)+sizeof(int);
	}

//--------------------------------------------------------------------------------------------
//TODO - use the newer functions available to discover the capabilities instead of hardcoding.
bool CNICard :: Set( int iCard, const char * pchName, const char * pchModel ) { 
	m_iCard=iCard;
	m_iIGains=1;
	m_afIGain[0]=1.0;
	m_iOGains=1;
	m_afOGain[0]=1.0;
	
	switch( iCard ) { 
		//-- M Series Input only devices 
		case 6220 : 
		case 6224 : 
			m_dInRange=10;
			m_dOutRange=10;

			m_iAIChannels = (iCard==6224 ? 32 : 16 );
			m_iAIBits = 16;	
			m_iMaxIRate = 1000000;
			m_iAOChannels = 0; 
			m_iAOBits = 16;	
			m_iMaxORate = 1;
			m_iDChannels = (iCard==6220?24:48);
			
			m_iIGains=4;
			m_afIGain[1]=2.0;
			m_afIGain[2]=10.0;
			m_afIGain[3]=50.0;
			break;

		//-- M Series Input/Output devices 
		case 6221 : 
		case 6225 : 
		case 6229 :
			m_dInRange=10;
			m_dOutRange=10;

			m_iAIChannels = (iCard==6225 ? 80 : 16 );
			m_iAIBits = 16;	
			m_iMaxIRate = 1000000;
			m_iAOChannels = (iCard==6229 ? 4 : 2 ); 
			m_iAOBits = 16;	
			m_iMaxORate = 833000;
			m_iDChannels = (iCard==6229 ? 48 : (iCard==6221 ? 10 : 24 ));

			m_iIGains=4;
			m_afIGain[1]=2.0;
			m_afIGain[2]=10.0;
			m_afIGain[3]=50.0;
			//TODO - these cards have settable output gains
			break;
		
		default : 
			ErrorMessage(IDS_NICARDUNRECOGNIZED);
			return false;
		}
	
	//---- copy provided name and model, up to NI_NAMESIZE chars.
	memcpy( m_achName, pchName, __min(strlen(pchName)+1, NI_NAMESIZE ) );
	m_achName[NI_NAMESIZE-1]=0;
	memcpy( m_achModel, pchModel, __min(strlen(pchModel)+1, NI_NAMESIZE ) );
	m_achName[NI_NAMESIZE-1]=0;
	return true;
	}

//-----------------------------------------------------------------------------
int NiCardFind( const char * pchName ) {
	for( int iCard=0; iCard<NiCardCount(); ++iCard ) 
		if( strcmp(g_aCards[iCard].m_achName, pchName)==0 ) 
			return iCard;
	return -1;
	}

//-----------------------------------------------------------------------------
bool NIInit() { 
	int iError;
	if( g_iCardCount >=0 ) 
		return false;
		
	char buf0[1024];					// list of card names from NI driver 
	char * pch, * pchCNo;

	//-- count the number of devices 
	g_iCardCount=0;
	if( iError=DAQmxGetSysDevNames(buf0, 1024) ) {
		ErrorMessage( IDS_NIGETDEVNAMES, iError ); 
		return false;
		}
	if( pch=strtok( buf0, "," ) ) 
		do { 
			++g_iCardCount;
			} while( pch=strtok( NULL, "," ) );
	if( g_iCardCount==0 ) 
		return true;
	
	//-- allocate a card array
	g_aCards = new CNICard[g_iCardCount];
	if( g_aCards==NULL ) { 
		ErrorMessage(IDS_NICARDALLOC);
		return false;
		}
	
	//-- Fill the cardarray 
	char buf1[1024];					// card name from NI 
	int iDev=0;
	if( pch=strtok( buf0, "," ) ) 
		do { 
			if( iError=DAQmxGetDevProductType(pch, buf1, 1024) )  {
				ErrorMessage( IDS_NIERRORCODE, iError ); 
				return false;
				}
			pchCNo=buf1; 
			while( *pchCNo!=0 && (*pchCNo<'0' || *pchCNo>'9' ) )
				++pchCNo;
			g_aCards[iDev].Set( atoi(pchCNo), pch, buf1 );
			} while( (pch=strtok( NULL, "," ))!=NULL && (++iDev<g_iCardCount) );

	//--------------------------------------------------
	#ifdef _DEBUG

	//------------------
	/*
	Results : 
	DAQmxGetDevAIPhysicalChans: Dev1/ai0, Dev1/ai1, Dev1/ai2, Dev1/ai3, Dev1/ai4, Dev1/ai5, Dev1/ai6, Dev1/ai7, Dev1/ai8, Dev1/ai9, Dev1/ai10, Dev1/ai11, Dev1/ai12, Dev1/ai13, Dev1/ai14, Dev1/ai15, Dev1/ai16, Dev1/ai17, Dev1/ai18, Dev1/ai19, Dev1/ai20, Dev1/ai21, Dev1/ai22, Dev1/ai23, Dev1/ai24, Dev1/ai25, Dev1/ai26, Dev1/ai27, Dev1/ai28, Dev1/ai29, Dev1/ai30, Dev1/ai31
	DAQmxGetDevAOPhysicalChans: Dev1/ao0, Dev1/ao1, Dev1/ao2, Dev1/ao3
	DAQmxGetDevDILines: Dev1/port0/line0, Dev1/port0/line1, Dev1/port0/line2, Dev1/port0/line3, Dev1/port0/line4, Dev1/port0/line5, Dev1/port0/line6, Dev1/port0/line7, Dev1/port0/line8, Dev1/port0/line9, Dev1/port0/line10, Dev1/port0/line11, Dev1/port0/line12, Dev1/port0/line13, Dev1/port0/line14, Dev1/port0/line15, Dev1/port0/line16, Dev1/port0/line17, Dev1/port0/line18, Dev1/port0/line19, Dev1/port0/line20, Dev1/port0/line21, Dev1/port0/line22, Dev1/port0/line23, Dev1/port0/line24, Dev1/port0/line25, Dev1/port0/line26, Dev1/port0/line27, Dev1/port0/line28, Dev1/port0/line29, Dev1/port0/line30, Dev1/port0/line31, Dev1/port1/line0, Dev1/port1/line1, Dev1/port1/line2, Dev1/port1/line3, Dev1/port1/line4, Dev1/port1/line5, Dev1/port1/line6, Dev1/port1/line7, Dev1/port2/line0, Dev1/port2/line1, Dev1/port2/line2, Dev1/port2/line3, Dev1/port2/line4, Dev1/port2/line5, Dev1/port2/line6, Dev1/port2/line7
	DAQmxGetDevDIPorts: Dev1/port0, Dev1/port1, Dev1/port2
	DAQmxGetDevDOLines: Dev1/port0/line0, Dev1/port0/line1, Dev1/port0/line2, Dev1/port0/line3, Dev1/port0/line4, Dev1/port0/line5, Dev1/port0/line6, Dev1/port0/line7, Dev1/port0/line8, Dev1/port0/line9, Dev1/port0/line10, Dev1/port0/line11, Dev1/port0/line12, Dev1/port0/line13, Dev1/port0/line14, Dev1/port0/line15, Dev1/port0/line16, Dev1/port0/line17, Dev1/port0/line18, Dev1/port0/line19, Dev1/port0/line20, Dev1/port0/line21, Dev1/port0/line22, Dev1/port0/line23, Dev1/port0/line24, Dev1/port0/line25, Dev1/port0/line26, Dev1/port0/line27, Dev1/port0/line28, Dev1/port0/line29, Dev1/port0/line30, Dev1/port0/line31, Dev1/port1/line0, Dev1/port1/line1, Dev1/port1/line2, Dev1/port1/line3, Dev1/port1/line4, Dev1/port1/line5, Dev1/port1/line6, Dev1/port1/line7, Dev1/port2/line0, Dev1/port2/line1, Dev1/port2/line2, Dev1/port2/line3, Dev1/port2/line4, Dev1/port2/line5, Dev1/port2/line6, Dev1/port2/line7
	DAQmxGetDevDOPorts: Dev1/port0, Dev1/port1, Dev1/port2
	*/
	/*
	char buf[2048];
	DAQmxGetDevAIPhysicalChans(g_aCards[0].m_achName, buf, 2048);
	OutputDebugString( "DAQmxGetDevAIPhysicalChans:" );
	OutputDebugString( buf );
	OutputDebugString( "\n" );
	DAQmxGetDevAOPhysicalChans(g_aCards[0].m_achName, buf, 2048);
	OutputDebugString( "DAQmxGetDevAOPhysicalChans:" );
	OutputDebugString( buf );
	OutputDebugString( "\n" );
	DAQmxGetDevDILines(g_aCards[0].m_achName, buf, 2048);
	OutputDebugString( "DAQmxGetDevDILines:" );
	OutputDebugString( buf );
	OutputDebugString( "\n" );
	DAQmxGetDevDIPorts(g_aCards[0].m_achName, buf, 2048);
	OutputDebugString( "DAQmxGetDevDIPorts:" );
	OutputDebugString( buf );
	OutputDebugString( "\n" );
	DAQmxGetDevDOLines(g_aCards[0].m_achName, buf, 2048);
	OutputDebugString( "DAQmxGetDevDOLines:" );
	OutputDebugString( buf );
	OutputDebugString( "\n" );
	DAQmxGetDevDOPorts(g_aCards[0].m_achName, buf, 2048);
	OutputDebugString( "DAQmxGetDevDOPorts:" );
	OutputDebugString( buf );
	OutputDebugString( "\n" );
	*/
	//------------------

	//------------------
	/*
	float64 aGain[16];
	
	DAQmxGetDevAIGains( g_aCards[0].m_achName, aGain, 16 );
	sprintf( buf, "AI Gains{%f,%f,%f,%f,%f,...}\n", aGain[0], aGain[1], aGain[2], aGain[3], aGain[4] );
	OutputDebugString( buf );

	DAQmxGetDevAOGains( g_aCards[0].m_achName, aGain, 16 );
	sprintf( buf, "AO Gains{%f,%f,%f,%f,%f,...}\n", aGain[0], aGain[1], aGain[2], aGain[3], aGain[4] );
	OutputDebugString( buf );
	*/
	
	#endif
	
	return true;
	}
	
//----------------------------------------------------------------------------
// create a new task 
CTask :: CTask( CNICard * poCard, 
				bool bAcq, int iAcqSampling, int iAcqBufferSize, 
				bool bGen, int iGenSampling, int iGenBufferSize,
				bool bFeedback, 
				bool bDigAcq, 
				bool bDigGen ) { 
	/*
	#ifdef _DEBUG
	char bufODS[1024];
	sprintf( bufODS, "NIDAQ CTASK( *, Acq=%c, %d, %d, Gen=%c, %d, %d, FdBk=%c, DAcq=%c, DGen=%c)\n", 
				bAcq?'Y':'N', iAcqSampling, iAcqBufferSize, 
				bGen?'Y':'N', iGenSampling, iGenBufferSize,
				bFeedback?'Y':'N', 
				bDigAcq?'Y':'N', 
				bDigGen?'Y':'N' );
	OutputDebugString( bufODS );
	#endif
	*/
	m_poCard=poCard;
	m_iError=0;
	m_hDigGen=m_hDigAcq=m_hAnaAcq=m_hAnaGen=NULL;
	m_iDigAcqLines=m_iDigGenLines=0;
	m_iAcqChannels=m_iGenChannels=0;
	m_b100khz=false;

	//-- Acquisition	
	m_iAcqSampling = iAcqSampling;
	m_iAcqBufferSize = iAcqBufferSize;
	if( m_bAnaAcqTask=bAcq )					// Analog
		if( (m_iError=DAQmxCreateTask( NULL, &m_hAnaAcq )) != 0 ) { 
			ErrorMessage( IDS_NICREATETASKACQ, m_iError );
			return;
			}
	if( m_bDigAcqTask=bDigAcq )				// digital 
		if( (m_iError=DAQmxCreateTask( NULL, &m_hDigAcq )) != 0 ) { 
			ErrorMessage( IDS_NICREATETASKACQ, m_iError );
			return;
			}
	
	//-- Generation
	m_iGenSampling = iGenSampling;	
	m_iGenBufferSize = bFeedback ? 1 : iGenBufferSize;
	if( m_bAnaGenTask=bGen )					// Analog
		if( (m_iError=DAQmxCreateTask( NULL, &m_hAnaGen )) != 0 ) {
			ErrorMessage( IDS_NICREATETASKGEN, m_iError );
			return;
			}
	if( m_bDigGenTask=bDigGen )				// Digital 
		if( (m_iError=DAQmxCreateTask( NULL, &m_hDigGen )) != 0 ) {
			ErrorMessage( IDS_NICREATETASKGEN, m_iError );
			return;
			}
	}

//------------------------------------------------- Add Channels		
//----- Analog Input
bool CTask :: AddInputChannel( int iCh, double dGain, int iMode ) { 
	/*
	#ifdef _DEBUG
	char bufODS[1024];
	sprintf( bufODS, "NIDAQ AddInputChannel(%d,%f)\n", iCh, dGain );
	OutputDebugString( bufODS );
	#endif
	*/

	++m_iAcqChannels;
	char buffer[64];
	sprintf( buffer, "%s/ai%d", m_poCard->m_achName, iCh );
	m_iError = DAQmxCreateAIVoltageChan( m_hAnaAcq, buffer, NULL, iMode, 
					-1.0 * m_poCard->m_dInRange/dGain, m_poCard->m_dInRange/dGain, 
					DAQmx_Val_Volts, NULL );
	if( m_iError!=0 ) 
		ErrorMessage( IDS_NICREATECHANACQ, m_iError, buffer );
	return m_iError==0;
	}
		
//-----
bool CTask :: AddOutputChannel( int iCh, double dGain ) { 
	/*
	#ifdef _DEBUG
	char bufODS[1024];
	sprintf( bufODS, "NIDAQ AddOutputChannel(%d,%f)\n", iCh, dGain );
	OutputDebugString( bufODS );
	#endif
	*/
	++m_iGenChannels;
	char buffer[64];
	sprintf( buffer, "%s/ao%d", m_poCard->m_achName, iCh );
	m_iError = DAQmxCreateAOVoltageChan( m_hAnaGen, buffer, NULL, 
					-1.0 * m_poCard->m_dOutRange/dGain, m_poCard->m_dOutRange/dGain, 
					DAQmx_Val_Volts, NULL );
	if( m_iError!=0 ) 
		ErrorMessage( IDS_NICREATECHANGEN, m_iError, buffer );
	return m_iError==0;
	}
			
//-----
bool CTask :: AddDigitalOutputChannels( int iPort, int iLines, BYTE * aiLines ) { 
	/*
	#ifdef _DEBUG
	char bufODS[1024];
	sprintf( bufODS, "NIDAQ AddDigitalOutputChannel(%d,%d,{%d,...})\n", iPort, iLines, int(aiLines[0]) );
	OutputDebugString( bufODS );
	#endif
	*/
	
	m_iDigGenLines=iLines;
	char * buffer = new char[iLines*(32+strlen(m_poCard->m_achName))];
	buffer[0]=0;

	for( int iLine=0; iLine<iLines; ++iLine )
		sprintf( buffer+strlen(buffer), ",%s/port%d/line%d", m_poCard->m_achName, iPort, int(aiLines[iLine]) );
		
	m_iError = DAQmxCreateDOChan( m_hDigGen, buffer+1, NULL, DAQmx_Val_ChanForAllLines );	// DAQmx_Val_ChanPerLine
	if( m_iError!=0 ) 
		ErrorMessage( IDS_NICREATECHANDIGGEN, m_iError, buffer );
	
	delete [] buffer;
	return m_iError==0;
	}

//-----
bool CTask :: AddDigitalInputChannels( int iPort, int iLines, BYTE * aiLines ) { 
	/*
	#ifdef _DEBUG
	char bufODS[1024];
	sprintf( bufODS, "NIDAQ AddDigitalInputChannel(%d,%d,{%d,...})\n", iPort, iLines, int(aiLines[0]) );
	OutputDebugString( bufODS );
	#endif
	*/
	
	m_iDigAcqLines=iLines;
	char * buffer = new char[iLines*(32+strlen(m_poCard->m_achName))];
	buffer[0]=0;

	for( int iLine=0; iLine<iLines; ++iLine )
		sprintf( buffer+strlen(buffer), ",%s/port%d/line%d", m_poCard->m_achName, iPort, int(aiLines[iLine]) );
		
	m_iError = DAQmxCreateDIChan( m_hDigAcq, buffer+1, NULL, DAQmx_Val_ChanForAllLines );	// DAQmx_Val_ChanPerLine
	if( m_iError!=0 ) 
		ErrorMessage( IDS_NICREATECHANDIGACQ, m_iError, buffer );
	delete [] buffer;
	return m_iError==0;
	}

			
//------------------------------------------------- Start Task helpers 			
//-----
bool CTask :: SetRegenMode( TaskHandle h, bool bAllowRegen ) {
	if( (m_iError=DAQmxSetWriteRegenMode( h, bAllowRegen ? DAQmx_Val_AllowRegen : DAQmx_Val_DoNotAllowRegen ) ) !=0 ){
		ErrorMessage( IDS_NIWRITEREGEN , m_iError );
		return false;
		}
	return true;
	}
	
//-----
bool CTask :: SetOutputBuffer( TaskHandle h, int iSize ) {
	if( (m_iError=DAQmxCfgOutputBuffer( h, iSize ) ) != 0 ){
		ErrorMessage( IDS_NICFGOUTPUTBUFFER, m_iError );
		return false;
		}
	return true;
	}
	
bool CTask :: SetInputBuffer( TaskHandle h, int iSize ) {
	//-- override ai buffer size // todo - is this relevant ? 		
	if( (m_iError=DAQmxCfgInputBuffer( h, iSize ) ) != 0 ) {
		ErrorMessage( IDS_NICFGINPUTBUFFER, m_iError );
		return false;
		}
	return true;
	}

bool CTask :: SetTiming( TaskHandle h, int iRate, int iSize, char * pchTimingSource, bool bGenFinite ) { 
	//-- Set timing to onboard clock
	if( (m_iError = DAQmxCfgSampClkTiming( h, pchTimingSource, iRate, DAQmx_Val_Rising, 
				bGenFinite?DAQmx_Val_FiniteSamps:DAQmx_Val_ContSamps, 
				iSize) ) != 0 ) {
		ErrorMessage( IDS_NICFGSAMPCLKGEN , m_iError );
		return false;
		}
	return true;
	}

bool CTask :: PreloadGenData( TaskHandle hAna, TaskHandle hDig, GenCallBack pFn, void * pSession ) { 
	/*
	#ifdef _DEBUG
	char bufODS[1024];
	sprintf( bufODS, "NIDAQ PreloadGenData(...)\n" );
	OutputDebugString( bufODS );
	#endif
	*/
	
	double * adBuffer = NULL;
	int * aiBuffer = NULL; 
	
	int32 iDigWritten=-1, iAnaWritten=-1;
	bool bError=false;

	//---- Preload some data 
	if( m_bAnaGenTask ) 
		adBuffer = new double[m_iGenBufferSize*m_iGenChannels];
	if( m_bDigGenTask ) 
		aiBuffer = new int[m_iGenBufferSize];
	
	pFn( m_iGenBufferSize, adBuffer, aiBuffer, pSession, -1 );

	if( m_bAnaGenTask ) {
		m_iError = DAQmxWriteAnalogF64( hAna, m_iGenBufferSize, false, 0, 
						DAQmx_Val_GroupByChannel, (double *)adBuffer, &iAnaWritten, NULL ); 
		if( m_iError!=0 ) {
			ErrorMessage( IDS_NIWRITEANALOGINITIAL , m_iError );
			bError=true;
			}
		delete [] adBuffer;
		}

	if( m_bDigGenTask ) {
		m_iError = DAQmxWriteDigitalU32( hDig, m_iGenBufferSize, false, 0, 
						DAQmx_Val_GroupByChannel, (uInt32 *)aiBuffer, &iDigWritten, NULL ); 
		if( m_iError!=0 ) {
			ErrorMessage( IDS_NIWRITEDIGITALINITIAL, m_iError );
			bError=true;
			}
		delete [] aiBuffer;
		}

	//-- rewind the counters if not all data could be written - should not happen ? 
	if( ( iDigWritten>0 && iDigWritten<m_iGenBufferSize) 
		|| ( iAnaWritten>0 && iAnaWritten<m_iGenBufferSize) ) 
		pFn( min(iAnaWritten,iDigWritten)-m_iGenBufferSize, NULL, NULL, pSession, -1 );

	return true;
	}

bool CTask :: StartTasks( NIPOLLER_INIT * pInit, bool bGenFinite ) { 
	/*
	#ifdef _DEBUG
	char bufODS[1024];
	sprintf( bufODS, "NIDAQ StartTasks(%c,%c,%c,%c)\n",
			m_bAnaGenTask ? 'Y' : 'N', m_bDigGenTask ? 'Y' : 'N',
			m_bAnaAcqTask ? 'Y' : 'N', m_bDigAcqTask ? 'Y' : 'N');
	OutputDebugString( bufODS );
	#endif
	*/
	
	char achClockAO[64];
	char achClockAI[64];
	sprintf( achClockAO, "/%s/ao/SampleClock", m_poCard->m_achName ); 		
	sprintf( achClockAI, "/%s/ai/SampleClock", m_poCard->m_achName ); 		

	m_bGenFinite=bGenFinite;
	pInit->m_pTask=this;
	bool bFeedback = ( ((CSession *)pInit->m_pSession)->m_pFeedback!=NULL );

	//-- Setup Analog generation 
	if( ! bFeedback && m_bAnaGenTask ) { 
		if( ! SetRegenMode( m_hAnaGen, false )
			|| ! SetOutputBuffer( m_hAnaGen, m_iGenBufferSize ) 
			|| ! SetTiming( m_hAnaGen, m_iGenSampling, m_iGenBufferSize, NULL, m_bGenFinite ) 
			) 
			return false;
		}

	//-- Setup analog acquisition 
	if( m_bAnaAcqTask ) { 
		if( ! SetTiming( m_hAnaAcq, m_iAcqSampling, m_iAcqBufferSize, ( ! bFeedback && m_bAnaGenTask ) ? achClockAO : NULL ) 
			|| ! SetInputBuffer( m_hAnaAcq, m_iAcqBufferSize ) ) 
			return false;
		}

	//-- Setup digital generation 
	if( ! bFeedback && m_bDigGenTask ) { 
		//-- if digital only, assume an analog acq is running and use its clock.  
		// Prevent Digital without analog in the user interface.
		// TODO - Try to find a method to use the digital output only without analog clock.
			
		if( ! SetRegenMode( m_hDigGen, false )	// ( m_bAnaGenTask && ! SetRegenMode( m_hDigGen, false ) ) 
			|| ! SetOutputBuffer( m_hDigGen, m_iGenBufferSize ) )
			return false;

//		if( ! m_bAnaGenTask && ! m_bAnaAcqTask ) 
		if( m_b100khz ) { 
//			m_b100khz=true;
			//-- if no analog, use "100kHzTimebase" and set rate to 100000
			if( ! SetTiming( m_hDigGen, 100000, m_iGenBufferSize, "100kHzTimebase", m_bGenFinite ) )  {
				return false;
				}
			}
		else { 
//			m_b100khz=false;
			if( ! SetTiming( m_hDigGen, m_iGenSampling, m_iGenBufferSize, 
					( ! bFeedback && m_bAnaGenTask ) ? achClockAO : achClockAI, 
					m_bGenFinite )  ) 
				return false;
			}
		}

	if( ! bFeedback && ( m_bAnaGenTask || m_bDigGenTask ) ) 
		if( ! PreloadGenData( m_hAnaGen, m_hDigGen, pInit->m_pfnGenCallBack, pInit->m_pSession ) ) 
			return false;

	//-- Setup digital acquisition 	//TODO - if ! ( Gen or DigGen ) and Acq then use ai clock, not NULL ? 
	if( m_bDigAcqTask ) { 
		if( ! SetTiming( m_hDigAcq, m_iAcqSampling, m_iAcqBufferSize, 
				( ! bFeedback && m_bAnaGenTask ) ? achClockAO : (m_bAnaAcqTask?achClockAI:NULL) ) 
			|| ! SetInputBuffer( m_hDigAcq, m_iAcqBufferSize ) ) 
			return false;
		}

	//----- Create a polling thread 
	DWORD dwThreadID;
	if( (m_hPoller = CreateThread( NULL, 0, NIPoller, pInit, 0, &dwThreadID)) == NULL ) {
		ErrorMessage( IDS_POLLERTHREADCREATE );
		return false;
		}
	//SetThreadPriority( m_hPoller, THREAD_PRIORITY_ABOVE_NORMAL );
	
	return true;
	}

NIPOLLER_INIT * CreateInit( 
			AcqCallBack pfnAcqCallBack, 
			GenCallBack pfnGenCallBack, 
			TermCallBack pfnTermCallBack, 
			void * pSession ) { 
	NIPOLLER_INIT * pInit = new NIPOLLER_INIT;
	pInit->m_bCancelled=false;
	pInit->m_pfnAcqCallBack=pfnAcqCallBack;
	pInit->m_pfnGenCallBack=pfnGenCallBack;
	pInit->m_pfnTerminate=pfnTermCallBack;
	pInit->m_pSession=pSession;
	
	return pInit;
	}
	
//----- Naitonal Instruments MX series poller 
// delete buffers helper function 	
void NIPollerClearBuffers( double ** pAA, unsigned int ** pDA, double ** pAG, unsigned int ** pDG ) { 
	if( *pAA != NULL ) 
		delete [] *pAA;
	if( *pDA != NULL ) 
		delete [] *pDA;
	if( *pAG != NULL ) 
		delete [] *pAG;
	if( *pDG != NULL ) 
		delete [] *pDG;
	*pAA=NULL;
	*pDA=NULL;
	*pAG=NULL;
	*pDG=NULL;
	}
	
DWORD WINAPI NIPoller( LPVOID lpParam ) { 
	NIPOLLER_INIT * pInit = (NIPOLLER_INIT *) lpParam;

	uInt32 iAnaAcqSamples=0, iDigAcqSamples=0, iAnaGenSamples=0, iDigGenSamples=0;
	int32 iWritten;											// NI returns samples written here

	CTask * pTask = pInit->m_pTask;
	uInt32 iInCount=pTask->m_iAcqBufferSize;		// acquisition buffer size 
	uInt32 iOutCount=pTask->m_iGenBufferSize;		// generation buffer size 
	
	//---------- Create transfer buffers for transferring data between N.I. calls and CTask object.
	double * pAnaAcqBuf = NULL;
	unsigned int * pDigAcqBuf = NULL;
	double * pAnaGenBuf = NULL;
	unsigned int * pDigGenBuf = NULL;
	
	//----- Create transfer buffers
	if(	   ( iInCount >0 && pTask->m_bAnaAcqTask && (pAnaAcqBuf=new double[iInCount*pTask->m_iAcqChannels])==NULL )
		|| ( iInCount >0 && pTask->m_bDigAcqTask && (pDigAcqBuf=new unsigned int[iInCount])==NULL )
		|| ( iOutCount>0 && pTask->m_bAnaGenTask && (pAnaGenBuf=new double[max(2,iOutCount)*pTask->m_iGenChannels])==NULL )
		|| ( iOutCount>0 && pTask->m_bDigGenTask && (pDigGenBuf=new unsigned int[max(2,iOutCount)])==NULL ) 
		) {
		//-- exit if any errors
		ErrorMessage( IDS_GENERROR, 0, "Create poller transfer buffer" );
		NIPollerClearBuffers( &pAnaAcqBuf, &pDigAcqBuf, &pAnaGenBuf, &pDigGenBuf );
		return -1;
		}

	//---------- Start tasks ( between 1 and 4 concurrent tasks dependant on options )
	//----- Start the analog acq task
	int iError=0;
	uInt64 iAnalogInitSamp=0, iDigitalInitSamp=0;
	if( pAnaAcqBuf ) { 
		if( (iError = DAQmxStartTask( pTask->m_hAnaAcq )) != 0 ) {
			ErrorMessage( IDS_NISTARTTASKACQ , iError );
			NIPollerClearBuffers( &pAnaAcqBuf, &pDigAcqBuf, &pAnaGenBuf, &pDigGenBuf );
			return -5;
			}
		if( (iError = DAQmxGetReadTotalSampPerChanAcquired( pTask->m_hAnaAcq, &iAnalogInitSamp )) != 0 ) {
			#ifdef _DEBUG
			OutputDebugString("Error in analog DAQmxGetReadTotalSampPerChanAcquired()\n" );
			#endif
			iAnalogInitSamp=0;
			}
		}

	//----- Start the digital acq task 
	if( pDigAcqBuf ) { 
		if( (iError = DAQmxStartTask( pTask->m_hDigAcq )) != 0 ) {
			ErrorMessage( IDS_NISTARTTASKACQ, iError );
			NIPollerClearBuffers( &pAnaAcqBuf, &pDigAcqBuf, &pAnaGenBuf, &pDigGenBuf );
			return -6;
			}
		if( (iError = DAQmxGetReadTotalSampPerChanAcquired( pTask->m_hDigAcq, &iDigitalInitSamp )) != 0 ) {
			#ifdef _DEBUG
			OutputDebugString("Error in digital DAQmxGetReadTotalSampPerChanAcquired()\n" );
			#endif
			iDigitalInitSamp=0;
			}
		}
	//----- Start the analog generation task 
	if( pAnaGenBuf ) { 
		if( (iError = DAQmxStartTask( pTask->m_hAnaGen )) != 0 ) {
			ErrorMessage( IDS_NISTARTTASKGEN, iError );
			NIPollerClearBuffers( &pAnaAcqBuf, &pDigAcqBuf, &pAnaGenBuf, &pDigGenBuf );
			return -7;
			}
		}

	//----- Start the digital generation task 
	if( pDigGenBuf ) { 
		if( (iError = DAQmxStartTask( pTask->m_hDigGen )) != 0 ) {
			ErrorMessage( IDS_NISTARTTASKGEN, iError );
			NIPollerClearBuffers( &pAnaAcqBuf, &pDigAcqBuf, &pAnaGenBuf, &pDigGenBuf );
			return -8;
			}
		}

	//---------- Polling loop runs until cancelled 
	CFeedback * pFeedback = NULL ;
	if( pTask->m_bAnaGenTask )
		pFeedback = ((CSession *)pInit->m_pSession)->m_pFeedback;
	if( pFeedback != NULL  
		&& ( pFeedback->m_iInputIndex<0 
			|| pFeedback->m_iInputIndex>=pTask->m_iAcqChannels
			|| pFeedback->m_iOutputIndex<0 
			|| pFeedback->m_iOutputIndex>=pTask->m_iGenChannels ) 
			)
		pFeedback = NULL;
	
	// iMinAcqSamples : minimum samples to process - fill 10% of buffer before processing
	// bFirst : if Gen & Acq, ignore first sample - this flag indicates a sample is yet to be ignored.
	// dOut : double value being generated from feedback calculation
	bool bFirst=pTask->m_bAnaGenTask;		
	int iMinAcqSamples = (! bFirst || pFeedback==NULL ? pTask->m_iAcqSampling / 50 : 1 );	
	double dOut, dOutRange = pTask->m_poCard->m_dOutRange;

	while( ! (pInit->m_bCancelled) ) { 
		bool bActive=false;
		
		//-- Check digital and analog samples to be read 
		if( pTask->m_bAnaAcqTask ) { 
			//-- check how many samples are waiting to be read
			iError=DAQmxGetReadAvailSampPerChan( pTask->m_hAnaAcq, &iAnaAcqSamples );
			if( iError != 0 ) 
				ErrorMessage( IDS_NIPOLLERACQCOUNT, iError );
			//-- test for glitching
			if( iAnaAcqSamples > iInCount )		
				iAnaAcqSamples=iInCount;			// TODO - warning if this occurs during feedback calculation ?
			}

		if( pTask->m_bDigAcqTask ) { 
			//-- check how many samples are waiting to be read
			iError=DAQmxGetReadAvailSampPerChan( pTask->m_hDigAcq, &iDigAcqSamples );
			if( iError != 0 ) 
				ErrorMessage( IDS_NIPOLLERACQCOUNT, iError );
			//-- test for glitching
			if( iDigAcqSamples > iInCount )		
				iDigAcqSamples=iInCount;
			}
			
		//-- check for Analog acquisition data.  If enough waiting, process 
		if( ( ! pTask->m_bAnaAcqTask || int(iAnaAcqSamples)>iMinAcqSamples )
				&& ( ! pTask->m_bDigAcqTask || int(iDigAcqSamples)>iMinAcqSamples ) ) { 
			bActive=true;
			int iSamples = ! pTask->m_bAnaAcqTask 
							? iDigAcqSamples 
							: ! pTask->m_bDigAcqTask
								? iAnaAcqSamples 
								: min( iDigAcqSamples, iAnaAcqSamples );
								
			//-- Analog read		
			if( pTask->m_bAnaAcqTask ) { 	
				iError = DAQmxReadAnalogF64( pTask->m_hAnaAcq, iSamples, 0, 
									DAQmx_Val_GroupByScanNumber/*DAQmx_Val_GroupByChannel*/, 
									pAnaAcqBuf, iInCount*pTask->m_iAcqChannels, &iWritten, NULL ); 
				if( bFirst ) {
					// Error on 1st sample for gen+acq.  Copy 2nd sample to first (data not yet generated).
					memcpy( pAnaAcqBuf, pAnaAcqBuf+pTask->m_iAcqChannels, pTask->m_iAcqChannels*sizeof(double));
					if( pFeedback!=NULL ) 
						iMinAcqSamples = 1;
					}
				if( iError != 0 || iWritten != iSamples )
					ErrorMessage( IDS_NIPOLLERACQREAD, iError );
				}
		
			//-- Digital Read 
			if( pTask->m_bDigAcqTask ) { 
				iError = DAQmxReadDigitalU32( pTask->m_hDigAcq, iSamples, 0.0, 
									DAQmx_Val_GroupByScanNumber/*DAQmx_Val_GroupByChannel*/, 
									(uInt32 *)pDigAcqBuf, 
									iInCount, &iWritten, NULL ); 
				if( bFirst ) {	// Error on 1st sample for gen+acq.  Copy 2nd sample to first (data not yet generated).
					pDigAcqBuf[0]=pDigAcqBuf[1];
					bFirst=false;
					}
				if( iError != 0 || iWritten != iSamples )
					ErrorMessage( IDS_NIPOLLERACQREAD, iError );
				}
			
			//-- send to main thread 
			bFirst=false;
			if( ( pTask->m_bAnaAcqTask || pTask->m_bDigAcqTask ) && ! pInit->m_bCancelled ){ 
				//#ifdef _DEBUG
				//sprintf( bufODS, "NIPoller AcqCallBack( %d, {%f,...}, {%d,...%d,...},)\n", iSamples, pAnaAcqBuf[0], pDigAcqBuf[0], pDigAcqBuf[50] );
				//OutputDebugString( bufODS );
				//#endif
				pInit->m_pfnAcqCallBack( iSamples, pAnaAcqBuf, pDigAcqBuf, pInit->m_pSession );
				}
			}

		//-- refill generation buffer
		if( pFeedback!=NULL ) {				
			bActive=true;
			//-- software timed with Feedback 
			int iSample, iOutIndex;
			uInt64 iSamplesRead;
			iError = DAQmxGetReadTotalSampPerChanAcquired( pTask->m_hAnaAcq, &iSamplesRead );
			iSamplesRead -= iAnalogInitSamp;

			//-- get desired level
			pInit->m_pfnGenCallBack( 1, pAnaGenBuf, pDigGenBuf, pInit->m_pSession, int(iSamplesRead) );

			//-- feedback calculation
			iSample = (iAnaAcqSamples-1)*pTask->m_iAcqChannels+pFeedback->m_iInputIndex;
			iOutIndex = pFeedback->m_iOutputIndex;
			dOut = pFeedback->CalculateOutput( pAnaGenBuf[iOutIndex], pAnaAcqBuf[iSample], int(iSamplesRead) );

			//-- apply card limits
			pAnaGenBuf[iOutIndex]=min(dOutRange, max(-1.0*dOutRange, dOut ) );
			
			//-- write to card
			iError = DAQmxWriteAnalogF64( pTask->m_hAnaGen, 1, false, 0, 
									DAQmx_Val_GroupByChannel, pAnaGenBuf, &iWritten, NULL ); 

			//-- process errors
			if( iError != 0 || iWritten != 1 ) {
				char buffer[100];
				sprintf( buffer, "%d : %d w/ feedback, value=pAnaGenBuf[iOutIndex]", iError, iWritten, pAnaGenBuf[iOutIndex] );
				ErrorMessage( IDS_NIPOLLERGENWRITE, iError, buffer );
				}
			}
		else if( pTask->m_bGenFinite ) {	
			bActive=true;
			//-- hardware timed finite generation
			bool32 bDone;						
			iError = DAQmxIsTaskDone( pTask->m_bAnaGenTask ? pTask->m_hAnaGen : pTask->m_hDigGen, &bDone );
			if( iError != 0 )
				ErrorMessage( IDS_NIQUERYTASKCOMPLETE, iError );
			pInit->m_bCancelled |= (bDone!=0);
			if( ! pInit->m_bCancelled ) 
				Sleep(10);
			}

		else if( ! pInit->m_bCancelled ) {	//-- hardware timed continuous generation

			//-- Analog Generation
			if( pTask->m_bAnaGenTask ) { 
				iError=DAQmxGetWriteSpaceAvail( pTask->m_hAnaGen, &iAnaGenSamples );
				if( iError != 0 )
					ErrorMessage( IDS_NIPOLLERGENCOUNT, iError );
				else if( iAnaGenSamples>MINGENSAMPLES ) { 
					bActive=true;
					pInit->m_pfnGenCallBack( iAnaGenSamples, pAnaGenBuf, NULL, pInit->m_pSession, -1 );

					//-- limit range
					for( uInt32 ii=0; int(ii)<iAnaGenSamples*pTask->m_iGenChannels; ++ii ) {
						if( pAnaGenBuf[ii]>10 /*todo - consider ? pTask->m_dOutRange*/ ) 
							pAnaGenBuf[ii]=10 /*pTask->m_dOutRange*/;
						if( pAnaGenBuf[ii]<-10 /*pTask->m_dOutRange*-1*/ ) 
							pAnaGenBuf[ii]=-10 /*pTask->m_dOutRange*-1*/ ;
						}

					iError = DAQmxWriteAnalogF64( pTask->m_hAnaGen, iAnaGenSamples, false, 0, 
										DAQmx_Val_GroupByChannel, pAnaGenBuf, &iWritten, NULL ); 

					//-- process errors
					if( iError!=0 || iWritten!=iAnaGenSamples ) {
						char buffer[100];
						sprintf( buffer, "%d,%d,%d", iWritten, iAnaGenSamples, pTask->m_iGenChannels );
						ErrorMessage( IDS_NIPOLLERGENWRITE, iError, buffer );
						}
					}
				}

			//-- Digital Generation
			if( pTask->m_bDigGenTask ) { 
				iError=DAQmxGetWriteSpaceAvail( pTask->m_hDigGen, &iDigGenSamples );
				if( iError != 0 )
					ErrorMessage( IDS_NIPOLLERGENCOUNT, iError );
				else if( iDigGenSamples>MINGENSAMPLES ) { 
					bActive=true;
					pInit->m_pfnGenCallBack( iDigGenSamples, NULL, pDigGenBuf, pInit->m_pSession, -1 );
				
					//-- write digital to card
					iError = DAQmxWriteDigitalU32( pTask->m_hDigGen, iDigGenSamples, false, 0, 
									DAQmx_Val_GroupByChannel, (uInt32 *)pDigGenBuf, &iWritten, NULL ); 

					//-- process errors
					if( iError!=0 || iWritten!=iDigGenSamples ) {
						char buffer[100];
						sprintf( buffer, "%d,%d", iWritten, iDigGenSamples );
						ErrorMessage( IDS_NIPOLLERGENWRITE, iError, buffer );
						}
					}
				}
			}
			
		if( ! bActive ) 
			Sleep(20);
		}

	//#ifdef _DEBUG
	//OutputDebugString( "NIPoller Exit\n" );
	//#endif

	if( pInit->m_pfnTerminate != NULL ) 
		pInit->m_pfnTerminate( pInit->m_pSession );
	
	//----- exiting - cleanup
	int iCode=0;
	iError=0;
	
	//-- analog acq stop task
	if( pTask->m_bAnaAcqTask && (iCode=DAQmxStopTask(pTask->m_hAnaAcq)) != 0 )
		ErrorMessage( IDS_NISTOPACQ , iError=iCode );
	
	//-- digital acq stop task
	if( pTask->m_bDigAcqTask && (iCode=DAQmxStopTask(pTask->m_hDigAcq)) != 0 )
		ErrorMessage( IDS_NISTOPACQ , iError=iCode );
	
	//-- analog gen stop task
	if( pTask->m_bAnaGenTask ) { 
		if( (iCode=DAQmxStopTask(pTask->m_hAnaGen)) != 0 ) {
			if( iCode == 200010 ) 
				AddMessage( REPTYPE_WARNING, "Generation task stopped before completing" );
			else if( iCode != 0 )
				ErrorMessage( IDS_NISTOPGEN , iError=iCode );
			}
		
		//-- gen a 0 on all outputs
		memset( pAnaGenBuf, 0, sizeof(double)*pTask->m_iGenChannels*2 );	
		int iError1 = DAQmxCfgSampClkTiming( pTask->m_hAnaGen, NULL, pTask->m_iGenSampling, DAQmx_Val_Rising, DAQmx_Val_FiniteSamps, 2 );
		int iError2 = DAQmxWriteAnalogF64( pTask->m_hAnaGen, 2, true, 0, DAQmx_Val_GroupByChannel, pAnaGenBuf, &iWritten, NULL ); 
		iCode=DAQmxClearTask(pTask->m_hAnaGen);
		if( iCode != 0 )
			ErrorMessage( IDS_NISTOPGEN , iError=iCode );
		}
	
	//-- digital gen stop task
	if( pTask->m_bDigGenTask ) { 
		if( (iCode=DAQmxStopTask(pTask->m_hDigGen)) != 0 ) {
			if( iCode == 200010 ) 
				AddMessage( REPTYPE_WARNING, "Generation task stopped before completing" );
			else if( iCode != 0 )
				ErrorMessage( IDS_NISTOPGEN , iError=iCode );
			}

		//-- gen 0 on all lines 
		pDigGenBuf[0]=0;
		pDigGenBuf[1]=0;
		DAQmxCfgSampClkTiming( pTask->m_hDigGen, "100kHzTimebase", 100000, DAQmx_Val_Rising, DAQmx_Val_FiniteSamps, 2 );
		DAQmxWriteDigitalU32( pTask->m_hDigGen, 2, true, 0, DAQmx_Val_GroupByChannel, (uInt32 *)pDigGenBuf, &iWritten, NULL ); 
		Sleep(100);		
		DAQmxStopTask(pTask->m_hDigGen);

		iCode=DAQmxClearTask(pTask->m_hDigGen);
		if( iCode != 0 )
			ErrorMessage( IDS_NISTOPGEN , iError=iCode );

		}
	
	//--	
	NIPollerClearBuffers( &pAnaAcqBuf, &pDigAcqBuf, &pAnaGenBuf, &pDigGenBuf );
	if( pTask != NULL )
		delete pTask;
	delete pInit;
	
	return iError!=0 ? -9 : 0;
	}

/*
int DigOut32( int iVal, char * pchPorts ) { 
	int iErr=0;
	int32 iWritten;
	uInt32 digout[1];
	digout[0]=iVal;

	//-- Creat Task 
	TaskHandle hTask;
	if( (iErr=DAQmxCreateTask( NULL, &hTask )) != 0 ) 		
		return iErr;
	
	//-- Create channel			//DAQmx_Val_ChanPerLine
	if( iErr==0 ) 
		iErr = DAQmxCreateDOChan( hTask, pchPorts, NULL, DAQmx_Val_ChanForAllLines );
		
	//-- write data [1]
	if( iErr==0 ) 
		iErr = DAQmxWriteDigitalU32( hTask, 
						1,						//Samples 
						true,					//autostart		
						1.0,					//timeout
						DAQmx_Val_GroupByChannel //DAQmx_Val_GroupByScanNumber 
						digout,					//buffer 
						&iWritten, 
						NULL);					//Reserved 
					
	DAQmxClearTask( hTask );
	return iErr;
	}
*/

//----------------------------------------------------------------------------------
DWORD WINAPI DigitalOutputPoller( LPVOID lpParam ) { 
	if( lpParam==NULL ){
		OutputDebugString( "DOP QUIT\n" );
		return -2;
		}

	BYTE * pStatus = (BYTE *) lpParam;
	unsigned int * piReset = ((unsigned int *) lpParam)+1;
	unsigned int * piWrite = ((unsigned int *) lpParam)+2;
	unsigned int * piWritten = ((unsigned int *) lpParam)+3;

	//-- Initialize digital output task 
	TaskHandle hTask;
	int iErr=DAQmxCreateTask( NULL, &hTask );
	if( iErr != 0 ) 
		return iErr;
	iErr = DAQmxCreateDOChan( hTask, (char *)(pStatus+16), NULL, DAQmx_Val_ChanForAllLines );	// DAQmx_Val_ChanPerLine

	if( iErr != 0 ) {
		DAQmxClearTask( hTask );
		return iErr;
		}
		
	if( pStatus[1]==0 )				// no initial write
		*piWritten = *piWrite = 0;
	else 							// initial write - leave piWrite as specified but ensure it is not = to piWritten
		*piWritten = ~(*piWrite);
		
	//-- continuously poll for changes from main thread and write to port if any occur.  
	int32 iWritten;
	while( pStatus[2]==0 ) { 
		if( *piWritten != *piWrite ) { 
			//-- Write digital out.  Exit if an error occurs
			if( pStatus[0]==8 ) 
				iErr = DAQmxWriteDigitalU8( hTask, 
								1/*Samples*/, 
								true/*autostart*/, 
								1.0/*timeout*/, 
								DAQmx_Val_GroupByChannel /*DAQmx_Val_GroupByScanNumber*/, 
								(uInt8 *)(piWrite) /*buffer*/, 
								&iWritten, 
								NULL/*Reserved*/ );
			else {
				iErr = DAQmxWriteDigitalU32( hTask, 
								1/*Samples*/, 
								true/*autostart*/, 
								1.0/*timeout*/, 
								DAQmx_Val_GroupByChannel /*DAQmx_Val_GroupByScanNumber*/, 
								(uInt32 *)(piWrite) /*buffer*/, 
								&iWritten, 
								NULL/*Reserved*/ );
				}

			if( iErr != -200587 && ( iErr!=0 || iWritten!=1 ) ) {
				DAQmxClearTask( hTask );
				return iErr;
				}
				
			//-- copy value to 'last written' so it does not write again.
			*piWritten=*piWrite;	
			
			//-- cause another write if any autoreset are set, after slight delay 
			*piWrite &= ~(*piReset);
			/*
			char buf[100];
			sprintf( buf, "Written=%d Reset=%d Write=%d \n", *piWritten, *piReset, *piWrite );
			OutputDebugString( buf );								
			*/
			if( *piWritten != *piWrite )
				Sleep(1);
			}
		else 
			Sleep(50);					// nothing changing, take a short break.
		}

	// cleanup: close the valves
	if( pStatus[0]==8 ) {
				* (uInt8 *) piWrite = 0;
				iErr = DAQmxWriteDigitalU8( hTask, 
								1/*Samples*/, 
								true/*autostart*/, 
								1.0/*timeout*/, 
								DAQmx_Val_GroupByChannel /*DAQmx_Val_GroupByScanNumber*/, 
								(uInt8 *)(piWrite) /*buffer*/, 
								&iWritten, 
								NULL/*Reserved*/ );
	} else {
				* (uInt32 *) piWrite = 0;
				iErr = DAQmxWriteDigitalU32( hTask, 
								1/*Samples*/, 
								true/*autostart*/, 
								1.0/*timeout*/, 
								DAQmx_Val_GroupByChannel /*DAQmx_Val_GroupByScanNumber*/, 
								(uInt32 *)(piWrite) /*buffer*/, 
								&iWritten, 
								NULL/*Reserved*/ );
				}


	delete [] pStatus;		
	DAQmxClearTask( hTask );
	return 0;
	}
